
/* Simple program:  Create a blank window, wait for keypress, quit.

   Please see the SDL documentation for details on using the SDL API:
   /Developer/Documentation/SDL/docs.html
*/
   
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include <SDL/SDL.h>

#include <unistd.h>
#include <sys/time.h>

#ifdef WIN32
#include <windows.h>
#define usleep(usec) Sleep(usec/1000)
#endif

#define tick(timevar) {gettimeofday(&ru, NULL); timevar = (double)ru.tv_sec + ru.tv_usec / 1000000.0;}
#define tock(timevar, label) {gettimeofday(&ru, NULL); printf("%f seconds used %s\n", (double)ru.tv_sec + ru.tv_usec / 1000000.0 - timevar, label);}
#define tok(timevar, timeusedvar) {gettimeofday(&ru, NULL); timeusedvar = (double)ru.tv_sec + ru.tv_usec / 1000000.0 - timevar;}
#define tic() tick(starttime)
#define toc() tock(starttime, "")

//#include "model.h"
#include "model.cpp"

int main(int argc, char *argv[])
{
	Uint32 initflags = SDL_INIT_VIDEO;  /* See documentation for details */
	SDL_Surface *screen;
	int xsize = 640;
	int ysize = 480;
	
	Uint8  video_bpp = 32;
	Uint32 videoflags = SDL_SWSURFACE;
	int    done;
        SDL_Event event;

	/* Initialize the SDL library */
	if ( SDL_Init(initflags) < 0 ) {
		fprintf(stderr, "Couldn't initialize SDL: %s\n",
			SDL_GetError());
		exit(1);
	}
	float framerate = 30;
	int framestep = 0;
	int i;
	bool bench = false;
	bool jumpbench = false;
	int framefactor = 1;
	bool paused = false;
	int max_iter = 256;
	int vis_iter = 0;
	struct timeval ru; //real (time) usage
	double starttime;
	double drawtime = 0;
	double steptime = 0;
	double misctime = 0;
	double jumptime = 0;
	double temptime = 0;
	double framestart = 0;
	double frametime = 0;
	struct Position pos;
	pos.rc = -.5;
	pos.ic = 0;
	long double zoom = 1; // initialization of pos.rs and pos.ic needs xsize and ysize
	int copt;
	tick(starttime);
	while ((copt = getopt (argc, argv, "bjfr:i:z:x:y:")) != -1)
		switch (copt) {
			case 'a':
				break;
			case 'j':
				jumpbench = true;
			case 'b':
				bench = true;
				max_iter *= 2;
				framerate = 0;
				break;
			case 'f':
				videoflags = videoflags ^ SDL_FULLSCREEN;
				break;
			case 'r':
				pos.rc = (long double) atof(optarg);
				break;
			case 'i':
				pos.ic = -(long double) atof(optarg);
				break;
			case 'z':
				zoom = (long double) atof(optarg);
				break;
			case 'x':
				xsize = atoi(optarg);
				break;
			case 'y':
				ysize = atoi(optarg);
				break;
			case '?':
				if (optopt == 'x' || optopt == 'y')
					fprintf (stderr, "Option -%c requires an argument.\n", optopt);
				else if (isprint (optopt))
					fprintf (stderr, "Unknown option `-%c'.\n", optopt);
				else
					fprintf (stderr,
							 "Unknown option character `\\x%x'.\n",
							 optopt);
				return 1;
			default:
				abort ();
	}
	pos.is = 2.5*zoom;
	pos.rs = zoom*2.5*xsize/ysize;
#ifdef WIN32
	if (!(bench)) 	freopen("CON", "w", stdout); // this gets rid of the stdout.txt file
#endif
	/* Set video mode */
	screen=SDL_SetVideoMode(xsize, ysize, video_bpp, videoflags);
	if (screen == NULL) {
		ysize = 640;
		xsize = 480;
		screen=SDL_SetVideoMode(xsize, ysize, video_bpp, videoflags);
		if (screen == NULL) {
		fprintf(stderr, "Couldn't set 640x480x%d video mode: %s\n", video_bpp, SDL_GetError());}
		SDL_Quit();
		exit(2);
	}
	done = 0;
#ifdef WIN32
	MandelModel<long double> fractal = MandelModel<long double>::MandelModel(xsize, ysize, pos.rc, pos.ic, pos.rs, pos.is);
#else
	AbstractModel fractal = AbstractModel::AbstractModel(xsize, ysize, pos);
#endif
	if (jumpbench) {
#define toak1() {gettimeofday(&ru, NULL); jumptime += (double)ru.tv_sec + ru.tv_usec / 1000000.0 - starttime; tic();}
#define toak2() {gettimeofday(&ru, NULL); drawtime += (double)ru.tv_sec + ru.tv_usec / 1000000.0 - starttime; tic();}
#define toak3() {gettimeofday(&ru, NULL); misctime += (double)ru.tv_sec + ru.tv_usec / 1000000.0 - starttime; tic();}
		
									tic();		fractal.Jump(100);	toak1();	fractal.Draw(screen);	toak2();
		fractal.Pan (.20, 0);		toak3();	fractal.Jump(100);	toak1();	fractal.Draw(screen);	toak2();
		fractal.Zoom(.125);			toak3();	fractal.Jump(100);	toak1();	fractal.Draw(screen);	toak2();
		fractal.Pan (-.0625, .0625);toak3();	fractal.Jump(100);	toak1();	fractal.Draw(screen);	toak2();
		fractal.Zoom(.25);			toak3();	fractal.Jump(100);	toak1();	fractal.Draw(screen);	toak2();
		fractal.Pan ( .125, -.125);	toak3();	fractal.Jump(100);	toak1();	fractal.Draw(screen);	toak2();
		fractal.Zoom(.25);			toak3();	fractal.Jump(100);	toak1();	fractal.Draw(screen);	toak2();
		fractal.Pan (-.125, .00);	toak3();	fractal.Jump(100);	toak1();	fractal.Draw(screen);	toak2();
		fractal.Zoom(.125);			toak3();	fractal.Jump(100);	toak1();	fractal.Draw(screen);	toak2();
		fractal.Pan (-.25, .00);	toak3();	fractal.Jump(100);	toak1();	fractal.Draw(screen);	toak2();
		fractal.Zoom(.125);			toak3();	fractal.Jump(100);	toak1();	fractal.Draw(screen);	toak2();
		fractal.Pan (0, .25);		toak3();	fractal.Jump(100);	toak1();	fractal.Draw(screen);	toak2();
		fractal.Zoom(.125);			toak3();	fractal.Jump(100);	toak1();	fractal.Draw(screen);	toak2();
		fractal.Zoom(.125);			toak3();	fractal.Jump(200);	toak1();	fractal.Draw(screen);	toak2();
		fractal.Zoom(.125);			toak3();	fractal.Jump(200);	toak1();	fractal.Draw(screen);	toak2();
		
		printf("\r%2.3f draw, %2.3f jump, %2.3f misc, %2.3f sec total", drawtime, jumptime, misctime, drawtime+jumptime+misctime);
		
		//		for (i = 0; i<100; i++) {fractal.Step(); fractal.ReDraw(screen);}
		done = 1;
	}
//#endif
	while ( !done ) {
		if ((bench && fractal.GetCurIter() < max_iter) || (!paused || framestep)) {
			tok(starttime, temptime);
			misctime += temptime;
			tick(starttime);
			if ( !((!paused) && vis_iter > 16 && vis_iter < 10000 && fractal.GetCurIter() < vis_iter-4) )
				for (int n=0; n<framefactor; n++) fractal.Step();
			if (framestep) framestep--;
			tok(starttime, temptime);
			steptime += temptime;
			tick(starttime);
			fractal.ReDraw(screen);
			tok(starttime, temptime);
			drawtime += temptime;
			tick(starttime);
			if ((!paused) && vis_iter > 8 && vis_iter < 10000 && (fractal.GetCurIter() < vis_iter)) 
				fractal.Jump((vis_iter/8*7 < fractal.GetCurIter()) ? vis_iter : fractal.GetCurIter() + vis_iter/8);
			tok(starttime, temptime);
			jumptime += temptime;
			tick(starttime);
			if (bench)
				printf("\r%2.3f draw, %2.3f calc, %2.3f jump, %2.3f sec misc", drawtime, steptime, jumptime, misctime);
		}
		if (bench && fractal.GetCurIter() >= max_iter) done=true;
//#ifndef WIN32
		if (!bench && framerate) {
			tok(framestart, frametime);
			//printf(", %f fps", 1./frametime);
//#if HAVE_UNISTD_H
			if (frametime < 1./framerate) usleep((1./framerate - frametime)*1000000 / ((paused) ? 20 : 1));
//#endif
			tick(framestart);
		}
//#endif
		/* Check for events */
		while ( SDL_PollEvent(&event) ) {
			switch (event.type) {
				case SDL_MOUSEMOTION:
					break;
				case SDL_MOUSEBUTTONDOWN:
					break;
				case SDL_KEYDOWN:
					switch ( event.key.keysym.sym){
						case SDLK_UP:
							fractal.Pan(0,-1/8.);
							if (paused) fractal.CatchUp(); else fractal.Step(); 
							fractal.Draw(screen);
							break;
						case SDLK_DOWN:
							fractal.Pan(0,1/8.); 
							if (paused) fractal.CatchUp(); else fractal.Step(); 
							fractal.Draw(screen);
							break;
						case SDLK_RIGHT:
							fractal.Pan(1/8.,0);
							if (paused) fractal.CatchUp(); else fractal.Step(); 
							fractal.Draw(screen);
							break;
						case SDLK_LEFT:
							fractal.Pan(-1/8.,0);
							if (paused) fractal.CatchUp(); else fractal.Step(); 
							fractal.Draw(screen);
							break;
						case SDLK_EQUALS:
							i = fractal.GetCurIter();
							fractal.Zoom(.5);
							vis_iter = fractal.GetVisIter();
							if (paused) fractal.Jump(i); else fractal.Step(); 
							fractal.Draw(screen);
							break;
						case SDLK_MINUS:
							i = fractal.GetCurIter();
							fractal.Zoom(2.);
							vis_iter = fractal.GetVisIter();
							if (paused) fractal.Jump(i); else fractal.Step(); 
							fractal.Draw(screen);
							break;
						case SDLK_LEFTBRACKET:
							if (event.key.keysym.mod & (KMOD_LSHIFT | KMOD_RSHIFT)) {
								framefactor = (framefactor < 2) ? 1 : framefactor - 1;
							} else {
								if (framerate > 5) {
									framerate = (int)(((float)framerate / 1.5) + .5) ;
								} else if (framerate > 1) 
									framerate = framerate - 1;
							}
							break;
						case SDLK_RIGHTBRACKET:
							if (event.key.keysym.mod & (KMOD_LSHIFT | KMOD_RSHIFT)) {
								framefactor = framefactor + 1;
							} else {
								if (framerate > 5) {
									framerate = (int)(((float)framerate * 1.5) + .5) ;
								} else if (framerate > 0) 
									framerate = framerate + 1;
							}
							break;
/*						case SDLK_f:
							videoflags | SDL_FULLSCREEN;
							//SDL_Quit();
							//SDL_Init(initflags);
							screen=SDL_SetVideoMode(1280, 800, video_bpp, videoflags);
							fractal.SetPrecision('0', 1280, 800);
							//if (SDL_WM_ToggleFullScreen(screen) == 0) {
							if (screen==NULL) {
								printf("couldn't switch video modes\n");
								videoflags ^ SDL_FULLSCREEN;
								SDL_FreeSurface(screen);;
								screen=SDL_SetVideoMode(xsize, ysize, video_bpp, videoflags);
								fractal.Draw(screen);
							}								
							break;*/
						case SDLK_c:
							fractal.ChangeColors();
							fractal.Draw(screen);
							break;
						case SDLK_BACKSPACE:
							fractal.Clear();
							fractal.Step();
							fractal.Draw(screen);
							break;
						case SDLK_SPACE:
							paused = !paused;
							break;
						case SDLK_PERIOD:
							if (event.key.keysym.mod & (KMOD_LCTRL  | KMOD_RCTRL )) fractal.Jump(fractal.GetCurIter()+ 19);
							framestep++;
							break;
						case SDLK_COMMA:
							i = -1;
							if (event.key.keysym.mod & (KMOD_LSHIFT | KMOD_RSHIFT)) i *= 5;
							if (event.key.keysym.mod & (KMOD_LCTRL  | KMOD_RCTRL )) i *= 20;
							i = i + fractal.GetCurIter();
							fractal.Clear();
							fractal.Jump(i);
							fractal.Draw(screen);
							break;
						case SDLK_ESCAPE:
							done = 1;
							break;
						case SDLK_q:
							done = 1;
							break;
						default:
							break;
					}
					break;
				case SDL_QUIT:
					done = 1;
					break;
				default:
					break;
			}
		}
	}
	if (bench) printf("\n");
	/* Clean up the SDL library */
	SDL_Quit();
	return(0);
}
