/*
 *  model.cpp
 *  rima
 *
 *  Created by Jonathan Toomim on 8/2/08.
 *  Copyleft 2008. All rights reversed.
 *
 */
#define Min(a, b) ((a < b) ? a : b)
#define Max(a, b) ((a > b) ? a : b)
#define pi 3.14159265358979313
#include "model.h"
#include <math.h>

template <class vtype>
static inline vtype arctan2(vtype y, vtype x)
{
	vtype abs_y;
	vtype angle;
	if (y == 0.0)
		return (x<0)?0:0.5;
	
	abs_y = fabsf(y);
	
	/* If we are in quadrant II or III, flip things around */
	if (x < 0.0)
		angle = 3.0f - (x + abs_y)/(abs_y - x);
	else
		angle = 1.0f - (x - abs_y)/(abs_y + x);
	//angle *= 536870912.0;
	/* If we are in quadrant III or IV, negate to return an
	 answer in the range +-4 */
	if (y < 0.0)
		angle = -angle;
	angle = angle*.125+.5; // range: 0 to 1 (compared to -pi/2 to pi/2)
	return angle;
}

Uint32 blue_ice(SDL_PixelFormat *fmt, float h, float s, float v)
{   // this is some code that had a bug that produces pretty results
	float ha = h*6;
	int hi = (int)(ha);
	float f  = ha - floor(ha);
	v = v * 255;
	int v2mo = 0200112;//0211002;//0200112;
	int tmo  = 0420441;//0144024;//0420441;
	int pmo  = 0112200;//0002211;//0112200;
	int qmo  = 0041424;//0424140;//0041424;
	Uint32 v2 =(Uint32) (Uint8)v            <<(3+7&(v2mo>>hi*3));
	Uint32 p = (Uint32)(v * (1 - s))        <<(3+7&(pmo >>hi*3));
	Uint32 q = (Uint32)(v * (1 - f*s))      <<(3+7&(qmo >>hi*3));
	Uint32 t = (Uint32)(v * (1 - (1 - f)*s))<<(3+7&(tmo >>hi*3));
	return v2|p|q|t;
}

Uint32 circus(SDL_PixelFormat *fmt, float h, float s, float v)
{   // more buggy but pretty code
	float ha = h*6;
	int hi = (int)(ha);
	float f  = ha - floor(ha);
	v = v * 255;
	int v2mo = 0200112;//0211002;//0200112;
	int tmo  = 0420441;//0144024;//0420441;
	int pmo  = 0112200;//0002211;//0112200;
	int qmo  = 0412111;//0424140;//0041424;
	Uint32 v2 =(Uint32)(((Uint8)v)                   <<(((7&(v2mo>>hi*3))<<3)));
	Uint32 p = (Uint32)((Uint8)(v * (1 - s)))        <<(((7&(pmo >>hi*3))<<3));
	Uint32 q = (Uint32)((Uint8)(v * (1 - f*s)))      <<(((7&(qmo >>hi*3))<<3));
	Uint32 t = (Uint32)((Uint8)(v * (1 - (1 - f)*s)))<<(((7&(tmo >>hi*3))<<3));
	return v2 | p | q | t;
}

Uint32 yellow_custard(SDL_PixelFormat *fmt, float h, float s, float v)
{   // more buggy but pretty code
	float ha = h*6;
	int hi = (int)(ha);
	float f  = ha - floor(ha);
	v = v * 255;
	int v2mo = 0200112;//0211002;//0200112;
	int tmo  = 0120111;//0144024;//0420441;
	int pmo  = 0112200;//0002211;//0112200;
	int qmo  = 0011121;//0424140;//0041424;
	Uint32 v2 =(Uint32)(((Uint8)v)                   <<(((7&(v2mo>>hi*3))<<3)));
	Uint32 p = (Uint32)((Uint8)(v * (1 - s)))        <<(((7&(pmo >>hi*3))<<3));
	Uint32 q = (Uint32)((Uint8)(v * (1 - f*s)))      <<(((7&(qmo >>hi*3))<<3));
	Uint32 t = (Uint32)((Uint8)(v * (1 - (1 - f)*s)))<<(((7&(tmo >>hi*3))<<3));
	return v2 | p | q | t;
}

Uint32 rainbow_sherbet(SDL_PixelFormat *fmt, float h, float s, float v)
{   // more buggy but pretty code
	float ha = h*6;
	int hi = (int)(ha);
	float f  = ha - floor(ha);
	v = v * 255;
	int v2mo = 0200112;//0211002;//0200112;
	int tmo  = 0420441;//0144024;//0420441;
	int pmo  = 0112200;//0002211;//0112200;
	int qmo  = 0041424;//0424140;//0041424;
	Uint32 v2 =(Uint32)(((Uint8)v)                   <<(((7&(v2mo>>hi*3))<<3)));
	Uint32 p = (Uint32)((Uint8)(v * (1 - s)))        <<(((7&(pmo >>hi*3))<<3));
	Uint32 q = (Uint32)((Uint8)(v * (1 - f*s)))      <<(((7&(qmo >>hi*3))<<3));
	Uint32 t = (Uint32)((Uint8)(v * (1 - (1 - f)*s)))<<(((7&(tmo >>hi*3))<<3));
	return v2 | p | q | t;
}

Uint32 realHSVtoRGB(SDL_PixelFormat *fmt, float h, float s, float v)
{ // with the bug in rainbow_sherbet fixed; not as pretty
	float ha = h*6;
	int hi = (int)(ha);
	float f  = ha - floor(ha);
	v = v * 255;
	int v2mo = 0200112;//0211002;//0200112;
	int tmo  = 0323031;//0144024;//0420441;
	int pmo  = 0112200;//0002211;//0112200;
	int qmo  = 0031323;//0424140;//0041424;
	Uint32 v2 =((Uint32)((Uint8)v))                     <<((7&(v2mo>>hi*3))<<3);
	Uint32 p = ((Uint32)((Uint8)(v * (1 - s))))         <<((7&(pmo >>hi*3))<<3);
	Uint32 q = ((Uint32)((Uint8)(v * (1 - f*s))))       <<((7&(qmo >>hi*3))<<3);
	Uint32 t = ((Uint32)((Uint8)(v * (1 - (1 - f)*s)))) <<((7&(tmo >>hi*3))<<3);
	return v2 | p | q | t;
}
/*
Uint32 oldHSVtoRGB(SDL_PixelFormat *fmt, float h, float s, float v)
{ // uses switch/case instead of bitshifts and bitwise ors; slightly slower, i think
	float ha = h*6;
	int hi = (int)(ha);
	float f  = ha - floor(ha);
	v = v * 255;
	Uint8 v2 = (Uint8)v;
	Uint8 p = (Uint8)(v * (1 - s));
	Uint8 q = (Uint8)(v * (1 - f*s));
	Uint8 t = (Uint8)(v * (1 - (1 - f)*s));
	switch (hi) {
		case 0: return SDL_MapRGB(fmt, v2,t, p);
		case 1:	return SDL_MapRGB(fmt, q, v2,p);
		case 2:	return SDL_MapRGB(fmt, p, v2,t);
		case 3: return SDL_MapRGB(fmt, p, q, v2);
		case 4: return SDL_MapRGB(fmt, t, p, v2);
		case 5:	return SDL_MapRGB(fmt, v2,p, q);
		default:
		{return SDL_MapRGB(fmt, v2, v2, v2);}
	}
}*/

#ifdef __SSE2__
#ifdef __GNUC__
#define USE_SSE2
#endif
#endif

#ifdef USE_SSE2
//#ifndef WIN32
#include <xmmintrin.h>
typedef float v4sf __attribute__ ((vector_size (16)));
typedef int   v4si __attribute__ ((vector_size (16)));
union fvec {  v4sf v; float f[4];  }; // vlen == 4 (must be a constant)
union ivec {  v4si v; int   i[4];  };

static __inline__ v4sf arctan2v(v4sf y, v4sf x)
{
	//v4sf abs_y;
	v4sf angle;
	int i;
	//if not all in the same quadrant, switch to scalar
	if (1) {//fixme
		fvec ys;
		fvec xs;
		fvec angles;
		ys.v = y; xs.v = x;
		for (i=0;i<4;i++) {
			angles.f[i] = arctan2(ys.f[i], xs.f[i]);
		}
		return angles.v;
	} 
	return angle;
}
/*
inline v4si HSVtoRGBv(SDL_PixelFormat *fmt, v4sf h, v4sf s, v4sf v)
{
	v4sf ha = h*6;
	v4si hi = __builtin_ia32_ctvtps2pi(ha);
	if __builtin_ia32_cmpeqps((__m128)hi, 6); //fixme  
	v4sf f  = ha - __builtin_ia32_cvtpi2ps(ha);
	v = v * 255;
	v4si v2 = __builtin_ia32_ctvps2pi(v);
	v4si p = __builtin_ia32_ctvps2pi(v * (1 - s));
	Uint8 q = __builtin_ia32_ctvps2pi(v * (1 - f*s));
	Uint8 t = __builtin_ia32_ctvps2pi(v * (1 - (1 - f)*s));
	switch (hi) {
		case 0: return SDL_MapRGB(fmt, v2,t, p);
		case 1:	return SDL_MapRGB(fmt, q, v2,p);
		case 2:	return SDL_MapRGB(fmt, p, v2,t);
		case 3: return SDL_MapRGB(fmt, p, q, v2);
		case 4: return SDL_MapRGB(fmt, t, p, v2);
		case 5:	return SDL_MapRGB(fmt, v2,p, q);
		default:
		{return SDL_MapRGB(fmt, v2, v2, v2);}
	}
}*/
//#endif
#endif

AbstractModel::AbstractModel(int xsize, int ysize, Position newpos) {
	w = xsize;
	h = ysize;
	pos = newpos;
	prec = '0';
	fprec = 22;
	dprec = 52;
	lprec = 63;
	char newprec;
	newprec = this->GetPrecisionRequirement(w, h, pos);
	this->SetPrecision(newprec, w, h);
}

AbstractModel::~AbstractModel() {
	switch (prec) {
		case 'f':	delete fmodel; break;
		case 'd':	delete dmodel; break;
		case 'l':	delete lmodel; break;
}}		

char AbstractModel::GetPrecisionRequirement(int x, int y, struct Position newpos) {
	double minsize;
	minsize = Min(newpos.rs / newpos.rc / (long double)x, 
				  newpos.is / newpos.ic / (long double)y);
	if (minsize > pow(2, -fprec)) return 'f';
	if (minsize > pow(2, -dprec)) return 'd';
	if (minsize > pow(2, -lprec)) return 'l';
	else return 'l'; // eventually, return 'b' for bignum support
}

void AbstractModel::SetPrecision(char p) {this->SetPrecision(p, w, h);}
void AbstractModel::SetPrecision(char p, int xsize, int ysize) {
	pos = this->GetPos();
	//w=xsize;
	//h=ysize;
	if (p=='0') p=this->GetPrecisionRequirement(w, h, pos);
	if (p!=prec) {
		switch (prec) {
			case 'f':	delete fmodel; break;
			case 'd':	delete dmodel; break;
			case 'l':	delete lmodel; break;
		}
		switch (p) {
			case 'f':	fmodel = new MandelModel<float>      (w, h, pos.rc, pos.ic, pos.rs, pos.is); break;
			case 'd':	dmodel = new MandelModel<double>     (w, h, pos.rc, pos.ic, pos.rs, pos.is); break;
			case 'l':	lmodel = new MandelModel<long double>(w, h, pos.rc, pos.ic, pos.rs, pos.is); break;
		}
		prec = p;
	}
}

#define AbstractVoidDef(name_and_types_and_params, name_and_params) \
void AbstractModel::name_and_types_and_params { \
this->SetPrecision();\
switch (prec) { \
case 'f':	fmodel->name_and_params; break; \
case 'd':	dmodel->name_and_params; break; \
case 'l':	lmodel->name_and_params; break; \
}}

#define AbstractNonVoidDef(returntype, name_and_types_and_params, name_and_params) \
returntype AbstractModel::name_and_types_and_params { \
returntype returnval; \
switch (prec) { \
case 'f':	returnval = fmodel->name_and_params; break; \
case 'd':	returnval = dmodel->name_and_params; break; \
case 'l':	returnval = lmodel->name_and_params; break; \
} \
return returnval; \
}

AbstractVoidDef(Zoom(float zoomfac), Zoom(zoomfac))
AbstractVoidDef(Pan(float xpan, float ypan), Pan(xpan, ypan))
AbstractVoidDef(Clear(), Clear())
AbstractVoidDef(CatchUp(), CatchUp())
AbstractVoidDef(Jump(int target_iter), Jump(target_iter))
AbstractVoidDef(Draw(SDL_Surface* &surf), Draw(surf))
AbstractVoidDef(ReDraw(SDL_Surface* &surf), ReDraw(surf))
AbstractVoidDef(ChangeColors(), ChangeColors())
AbstractVoidDef(SetColors(int i), SetColors(i))
AbstractNonVoidDef(int, Step(), Step())
AbstractNonVoidDef(int, GetVisIter(int max_iter), GetVisIter(max_iter))
AbstractNonVoidDef(int, GetCurIter(), GetCurIter())
AbstractNonVoidDef(int, GetColorID(), GetColorID())
AbstractNonVoidDef(struct Position, GetPos(), GetPos())
		
/*	void	Zoom(float zoomfac);
	void	Pan(float xpan, float ypan);
	void	Clear();
	void	CatchUp();
	void	Jump(int target_iter);
	int 	Step();
	void	Draw(SDL_Surface* &surf);
	void	ReDraw(SDL_Surface* &surf);
	int		GetVisIter(int max_iter);
	int 	GetCurIter();
	void	ChangeColors();
	void	SetColors(int i);
	int		GetColorID();
 protected:
	int w; int h; // width and height
	struct Position pos;  real and imaginary center and size
};
void AbstractModel::Zoom(float zoomfac) {
	switch (prec) {
		case 'f':	pos = fmodel->Zoom(zoomfac); break;
		case 'd':	pos = dmodel->Zoom(zoomfac); break;
		case 'l':	pos = lmodel->Zoom(zoomfac); break;
	}}


void AbstractModel::Pan(float xpan, float ypan) {
	switch (prec) {
		case 'f':	pos = fmodel->Pan(xpan, ypan); break;
		case 'd':	pos = dmodel->Pan(xpan, ypan); break;
		case 'l':	pos = lmodel->Pan(xpan, ypan); break;
	}}

void AbstractModel::Clear()

struct Position AbstractModel::GetPos() {
	switch (prec) {
		case 'f':	pos = fmodel->GetPos(); break;
		case 'd':	pos = dmodel->GetPos(); break;
		case 'l':	pos = lmodel->GetPos(); break;
	}
	return pos;
}
*/

template <class vtype>
MandelModel<vtype>::MandelModel(int xsize, int ysize, long double rcenter, 
								long double icenter, long double rsize, 
								long double isize) 
{
	w  = xsize;	h = ysize;
	zr = new vtype[w*h]; 
	zi = new vtype[w*h];
	cr = new vtype[w];
	ci = new vtype[h];
	iter = new int[w*h];
	in_bounds = new bool[w*h];
	draw_this_pixel = new bool[w*h];
	screen_door_effect = false;
	rc = rcenter; ic = icenter; rs = rsize; is = isize;
	esc_rad = 2;
	esc_rad_sq = esc_rad*esc_rad;
	HSVtoRGBsize = 5;
	HSVtoRGBs = new colorconvtype[HSVtoRGBsize];
	HSVtoRGBs[0] = &rainbow_sherbet;
	HSVtoRGBs[1] = &circus;
	HSVtoRGBs[2] = &blue_ice;
	HSVtoRGBs[3] = &yellow_custard;
	HSVtoRGBs[4] = &realHSVtoRGB;
	//HSVtoRGBs[5] = &oldHSVtoRGB;
	HSVtoRGBindex = 0;
	HSVtoRGB = HSVtoRGBs[HSVtoRGBindex];
#ifdef USE_SSE2
	vlen = 16/sizeof(vtype);
	vin_bounds = new bool[w*h/vlen];
	voo_bounds = new bool[w*h/vlen];
	vdraw_this_pixel = new bool[w*h/vlen];
#endif	
	this->Clear();
}

template <class vtype>
MandelModel<vtype>::~MandelModel() {
	delete zr;
	delete zi;
	delete cr;
	delete ci;
	delete in_bounds;
	delete draw_this_pixel;
	delete iter;
	delete HSVtoRGBs;
#ifdef USE_SSE2
	delete vin_bounds;
	delete voo_bounds;
	delete vdraw_this_pixel;	
#endif
}

template <class vtype>
void MandelModel<vtype>::Zoom(float zoomfac) {
//	if (!(fmod(zoomfac*256, 1)) && zoomfac < .5) {this->Zoom(.5); return this->Zoom(zoomfac*2);} 
	rs = rs*zoomfac;
	is = is*zoomfac;

	if ((zoomfac == 2 || zoomfac == .5) && !(w % 4 || h % 4)) {
		int x=0; int xn=0; int y=0; int yn=0; int off=0; int offn=0;
		// we need to adjust in_bounds, draw_this_pixel, iter, zr, zi, cr, ci, 
		// cur_iter, catchup_iter, redraw_all, vin_bounds, voo_bounds
		vtype* zrtmp = new vtype[w/2*h/2]; // zr
		vtype* zitmp = new vtype[w/2*h/2]; // zi
		bool*  ibtmp = new bool [w/2*h/2]; // in_bounds
		int*   ittmp = new int  [w/2*h/2]; // iter
		//bool*  vitmp = new bool [w/2*h/2/vlen]; // vin_bounds // fixme!
		//bool*  votmp = new bool [w/2*h/2/vlen]; // voo_bounds

		for (y=0; y < h; y++) ci[y] = ic + is * vtype(y - h/2)/vtype(h);
		for (x=0; x < w; x++) cr[x] = rc + rs * vtype(x - w/2)/vtype(w);
		catchup_iter = cur_iter;
		cur_iter = 0;
		redraw_all = true;
		
		if (zoomfac == 2) { // zooming out
			for (y=0; y<h; y+=2) {
				yn = y/2;
				for (x=0; x<w; x+=2) {
					xn = x/2;
					off = y*w+x;
					offn = yn*w/2+xn;
					zrtmp[offn] = zr[off];
					zitmp[offn] = zi[off];
					ibtmp[offn] = in_bounds[off];
					ittmp[offn] = iter[off];
				}
			}
			memset( zr,				0,		sizeof(vtype)*w*h);
			memset( zi,				0,		sizeof(vtype)*w*h);
			memset( in_bounds,		true,	sizeof(bool) *w*h);
			memset( iter,			0,		sizeof(int)  *w*h);			
			memset( draw_this_pixel,true,	sizeof(bool) *w*h);
#ifdef USE_SSE2
			memset( vin_bounds,		true,	sizeof(bool) *w*h/vlen);
			memset( voo_bounds,		false,	sizeof(bool) *w*h/vlen);
			memset( vdraw_this_pixel,true,	sizeof(bool) *w*h/vlen);
#endif
			for (y=h/4; y<3*h/4; y++) {
				off = y*w + w/4;
				offn = (y-h/4)*w/2;
				memcpy( zr+off,			zrtmp+offn,	sizeof(vtype)*w/2);
				memcpy( zi+off,			zitmp+offn,	sizeof(vtype)*w/2);
				memcpy( in_bounds+off,	ibtmp+offn,	sizeof(bool) *w/2);
				memcpy( iter+off,		ittmp+offn,	sizeof(int)  *w/2);
			}
		} else {
			if (zoomfac == .5) { // zooming in
				for (y=h/4; y<3*h/4; y++) {
					off = y*w + w/4;
					offn = (y-h/4)*w/2;
					memcpy( zrtmp+offn,	zr+off,			sizeof(vtype)*w/2);
					memcpy( zitmp+offn,	zi+off,			sizeof(vtype)*w/2);
					memcpy( ibtmp+offn,	in_bounds+off,	sizeof(bool) *w/2);
					memcpy( ittmp+offn,	iter+off,		sizeof(int)  *w/2);
				}
				memset( zr,				0,		sizeof(vtype)*w*h);
				memset( zi,				0,		sizeof(vtype)*w*h);
				memset( in_bounds,		true,	sizeof(bool) *w*h);
				memset( iter,			0,		sizeof(int)  *w*h);			
				memset( draw_this_pixel,true,	sizeof(bool) *w*h);
#ifdef USE_SSE2
				memset( vin_bounds,		true,	sizeof(bool) *w*h/vlen);
				memset( voo_bounds,		false,	sizeof(bool) *w*h/vlen);
				memset( vdraw_this_pixel,true,	sizeof(bool) *w*h/vlen);
#endif
				for (y=0; y<h; y+=2) {
					yn = y/2;
					for (x=0; x<w; x+=2) {
						xn = x/2;
						off = y*w+x;
						offn = yn*w/2+xn;
						if (screen_door_effect || !ibtmp[offn]) 
						{
							zr[off] = zrtmp[offn];
							zi[off] = zitmp[offn];
							in_bounds[off] = ibtmp[offn];
							iter[off] = ittmp[offn];
						}
					}
				}
			}
		}
		delete zrtmp; delete zitmp; delete ibtmp; delete ittmp;
	}
	else this->Clear(); 
}

template <class vtype>
void MandelModel<vtype>::Pan(float xpan, float ypan) {
	ypan = float(int(ypan*h))/float(h);
	xpan = float(int(xpan*w))/float(w);
	rc = rc + rs * xpan;
	ic = ic + is * ypan;
	for (int y=0; y < h; y++) ci[y] = ic + is * vtype(y - h/2)/vtype(h);
	for (int x=0; x < w; x++) cr[x] = rc + rs * vtype(x - w/2)/vtype(w);
	int copysize = w*h - int(fabs(ypan)*h)*w - int(fabs(xpan)*w);
	int copysrc = Max(int( ypan*h)*w, 0) + Max(int( xpan*w), 0);
	int copydst = Max(int(-ypan*h)*w, 0) + Max(int(-xpan*w), 0);
	vtype* tmp = new vtype[w*h];
	bool*  tm2 = new bool [w*h];
	int*   tm3 = new int  [w*h];
	bool*  tm4 = new bool [w*h/vlen];
	memcpy( tmp+copydst,		zr	+copysrc,				sizeof(vtype)*copysize);
	memcpy( zr,					tmp,						sizeof(vtype)*w*h);
	memcpy( tmp+copydst,		zi	+copysrc,				sizeof(vtype)*copysize);
	memcpy( zi,					tmp,						sizeof(vtype)*w*h);
	memcpy( tm2+copydst,		in_bounds+copysrc,			sizeof(bool) *copysize);
	memcpy( in_bounds,			tm2,						sizeof(bool) *w*h);
	memcpy( tm3+copydst,		iter + copysrc,				sizeof(int)  *copysize);	
	memcpy( iter,				tm3,						sizeof(int)  *w*h);	
#ifdef USE_SSE2
	memcpy( tm4+copydst/vlen,	vin_bounds+copysrc/vlen,	sizeof(bool) *copysize/vlen);
	memcpy( vin_bounds,			tm4,						sizeof(bool) *w*h/vlen);
	memcpy( tm4+copydst/vlen,	voo_bounds+copysrc/vlen,	sizeof(bool) *copysize/vlen);
	memcpy( voo_bounds,			tm4,						sizeof(bool) *w*h/vlen);
#endif
	delete tmp; delete tm2; delete tm3; delete tm4;
	
	if (xpan) {
	  int blankstart = (xpan < 0) ? 0 : w-int(xpan*w);
	  int blanksize  = int(fabs(xpan)*float(w));
	  for (int y=0; y<h; y++) {
		memset( zr+blankstart+y*w,					0,		sizeof(vtype)*blanksize);
		memset( zi+blankstart+y*w,					0,		sizeof(vtype)*blanksize);
		memset( in_bounds+blankstart+y*w,			true,	sizeof(bool) *blanksize);
		memset( iter+blankstart+y*w,				0,		sizeof(int)  *blanksize);			
#ifdef USE_SSE2
		memset( vin_bounds+(blankstart+y*w)/vlen,	true,	sizeof(bool) *blanksize/vlen);
		memset( voo_bounds+(blankstart+y*w)/vlen,	false,	sizeof(bool) *blanksize/vlen);
#endif
	  }
	}
	if (ypan) {
		int blankstart = (ypan < 0) ? 0 : h*w-int(ypan*h)*w;
		int blanksize  = w*int(fabs(ypan)*float(h));
		memset( zr+blankstart,			0,		sizeof(vtype)*blanksize);
		memset( zi+blankstart,			0,		sizeof(vtype)*blanksize);
		memset( in_bounds+blankstart,	true,	sizeof(bool) *blanksize);
		memset( iter+blankstart,		0,		sizeof(int)  *blanksize);
#ifdef USE_SSE2
		memset( vin_bounds+blankstart/vlen,	true,	sizeof(bool) *blanksize/vlen);
		memset( voo_bounds+blankstart/vlen,	false,	sizeof(bool) *blanksize/vlen);
#endif
	}
	catchup_iter = (catchup_iter > cur_iter) ? catchup_iter : cur_iter;
	cur_iter = 0; 
	redraw_all = true;
	memset( draw_this_pixel, true, sizeof(bool) *h*w);
#ifdef USE_SSE2
	memset( vin_bounds,	true, sizeof(bool) *w*h/vlen);
	memset( voo_bounds,	false, sizeof(bool) *w*h/vlen);
	memset( vdraw_this_pixel,	true, sizeof(bool) *w*h/vlen);
#endif
	//this->Clear(); //FIXME
}

template <class vtype>
void MandelModel<vtype>::Clear() {
	cur_iter = catchup_iter = 0;
	memset( zr,			0, sizeof(vtype)*w*h);
	memset( zi,			0, sizeof(vtype)*w*h);
	memset( in_bounds,	true, sizeof(bool) *w*h);
	memset( draw_this_pixel,	true, sizeof(bool) *w*h);
	memset( iter,		0, sizeof(int)  *w*h);
#ifdef USE_SSE2
	memset( vin_bounds,	true, sizeof(bool) *w*h/vlen);
	memset( voo_bounds,	false, sizeof(bool) *w*h/vlen);
	memset( vdraw_this_pixel,	true, sizeof(bool) *w*h/vlen);
#endif
	int x; int y;
	for (y=0; y < h; y++) {
		ci[y] = ic + is * vtype(y - h/2)/vtype(h);
	}
	for (x=0; x < w; x++) {
		cr[x] = rc + rs * vtype(x - w/2 )/vtype(w );
	}
}

template <class vtype>
void MandelModel<vtype>::Jump(int target_iter) {
	int x; int y; int i; int off; int vi; const int step_size = 16;
	vi = Min(this->GetVisIter(10000), target_iter);
#pragma omp parallel for private(x, i, off) firstprivate(vi)
	for (y=0; y<h; y++) {
		for (x=0; x<w ; x++) {
			off = x+w*y;
			i=iter[off];
			vtype cx=cr[x]; vtype cy=ci[y]; 
			vtype tmp; vtype zit; vtype zrt; bool cont = true;
			zit = zi[off]; zrt = zr[off];
			if (in_bounds[off] && i < vi-step_size) {  // pre-jump
				while (cont && i<vi-step_size) {					
					tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
					tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
					tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
					tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
					
					tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
					tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
					tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
					tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
					
					zr[off] = zrt; zi[off] = zit; i = i + step_size; iter[off] = i;
				}
			}
			for (; in_bounds[off] && i<target_iter ; i++) {
				vtype zr2 = zrt*zrt;
				vtype zi2 = zit*zit;
				if (zi2+zr2 > esc_rad_sq) {
					in_bounds[off] = false;
					iter[off] = i;
					zi[off] = zit;
					zr[off] = zrt;
				} else {
				zit = 2*zrt*zit + cy;
				zrt = zr2-zi2 + cx;
				}
			}
			if (in_bounds[off]) {	iter[off]=i;	zi[off] = zit;	zr[off] = zrt;	}
		}
	}
	cur_iter = target_iter;
}

#if 0
template <class vtype>
void MandelModel<vtype>::Jump(int target_iter) {
	int x; int y; int i; int off;
#pragma omp parallel for private(x, i, off)  
	for (y=0; y<h; y++) {
		for (x=0; x<w ; x++) {
			off = x+w*y;
			for (i=iter[off]; in_bounds[off] && i<target_iter ; i++) {
				vtype zr2 = zr[off]*zr[off];
				vtype zi2 = zi[off]*zi[off];
				if (zi2+zr2 > esc_rad_sq) {
					in_bounds[off] = false;
					iter[off] = i;
				} else {
					zi[off] = 2*zr[off]*zi[off] + ci[y];
					zr[off] = zr2-zi2 + cr[x];
				}
			}
			if (in_bounds[off]) iter[off]=i;
		}
	}
	cur_iter = target_iter;
}
#endif

#ifdef __GNsfsdfUC__
template <class vtype>
void MandelModel<vtype>::Jump(int target_iter) {
	int x; int y; int i; int off; int vi; const int step_size = 16;
	vi = Min(this->GetVisIter(), target_iter);
#pragma omp parallel for private(x, i, off) firstprivate(vi)
	for (y=0; y<h; y++) {
		for (x=0; x<w ; x++) {
			off = x+w*y;
			i=iter[off];
			vtype cx=cr[x]; vtype cy=ci[y]; 
			vtype tmp; vtype zit; vtype zrt; bool cont = true;
			zit = zi[off]; zrt = zr[off];
			if (in_bounds[off] && i < vi-step_size) {  // pre-jump
				while (cont && i<vi-step_size) {					
					tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
					tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
					tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
					tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
					
					tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
					tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
					tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
					tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
					
					zr[off] = zrt; zi[off] = zit; i = i + step_size; iter[off] = i;
				}
			}
			for (; in_bounds[off] && i<target_iter ; i++) {
				vtype zr2 = zrt*zrt;
				vtype zi2 = zit*zit;
				if (zi2+zr2 > esc_rad_sq) {
					in_bounds[off] = false;
					iter[off] = i;
					zi[off] = zit;
					zr[off] = zrt;
				} else {
					zit = 2*zrt*zit + cy;
					zrt = zr2-zi2 + cx;
				}
			}
			if (in_bounds[off]) {	iter[off]=i;	zi[off] = zit;	zr[off] = zrt;	}
		}
	}
	cur_iter = target_iter;
}
#endif

template <class vtype>
int MandelModel<vtype>::Step() {
	int x; int y; int off;
	vtype zr2; vtype zi2;
#pragma omp parallel for private(x, zr2, zi2, off) schedule(static) 
	for (y=0; y<h; y++) {
		for (x=0; x<w ; x++) {
			off = x+y*w;
			if (in_bounds[off] && iter[off] <= cur_iter) {
				zr2 = zr[off]*zr[off];
				zi2 = zi[off]*zi[off];
				if (zi2+zr2 > esc_rad_sq) {
					in_bounds[off] = false;
				} else {
					zi[off] = 2*zr[off]*zi[off] + ci[y];
					zr[off] = zr2 - zi2 + cr[x];
					iter[off]++;// = iter[x+w*y]+1;
				}
			}
		}
	}
	catchup_iter = (catchup_iter > cur_iter) ? catchup_iter : 0;
	return cur_iter++;
}

#ifdef USE_SSsdfE2
template <>
int MandelModel<float>::Step() {
	typedef float vtype;
	int not_vectorized = 0; int vectorized=0;
	int vx; int y; int voff; int off; int i; int mask;
	v4si iterv; v4si cur_iterv = {cur_iter,cur_iter,cur_iter,cur_iter}; v4si onesv = {1,1,1,1}; 
	v4sf civ; fvec tmpv; //v4sf zr2v; v4sf zi2v; v4sf zrv; v4sf ziv;  v4sf crv;
	v4sf esc_rad_sqv = {esc_rad_sq,esc_rad_sq,esc_rad_sq,esc_rad_sq}; v4sf twosv = {2,2,2,2};
	vtype zr2; vtype zi2; 
#pragma omp parallel for private( vx, voff, off, i, mask, iterv, civ, tmpv, zr2, zi2 ) schedule(static) 

	for (y=0; y<h; y++) {
		for (i=0; i<vlen; i++) tmpv.f[i] = ci[y];
		civ = tmpv.v;
		for (vx=0; vx<w/vlen; vx++) {
			voff = vx*vlen+w*y;			
			if (!voo_bounds[voff/vlen]) {
				iterv = (v4si)_mm_load_ps((float*)&iter[voff]);
				if (vin_bounds[voff/vlen] && (!catchup_iter || 
					!(__builtin_ia32_pmovmskb128(__builtin_ia32_cmpgtps((v4sf)iterv, (v4sf)cur_iterv))))){
					vectorized++;
					v4sf zrv = _mm_load_ps(&zr[voff]);
					v4sf ziv = _mm_load_ps(&zi[voff]);
					v4sf zr2v = zrv*zrv;
					v4sf zi2v = ziv*ziv;
					if (__builtin_expect(mask = __builtin_ia32_pmovmskb128(__builtin_ia32_cmpgtps(zr2v+zi2v, esc_rad_sqv)), 0)) {
						vin_bounds[voff/vlen] = false;
						int noob = 0;
						for (i=0; i<vlen; i++) {
							not_vectorized++;
							if (mask & (15<<(4*i))) {
								in_bounds[voff+i] = false;
								noob++;
							} else {
								zr2 = zr[voff+i]*zr[voff+i];
								zi2 = zi[voff+i]*zi[voff+i];
								zi[voff+i] = 2*zi[voff+i]*zr[voff+i] + ci[y];
								zr[voff+i] = zr2 - zi2 + cr[vx*vlen+i];
								iter[voff+i]++;
							}
						}
						if (noob==vlen) voo_bounds[voff/vlen] = true;
					} else {
						v4sf crv =  _mm_load_ps(&cr[vx*vlen]);
						_mm_store_ps(&zi[voff], twosv*ziv*zrv + civ);
						_mm_store_ps(&zr[voff], zr2v - zi2v + crv);
						_mm_store_ps((float*)&iter[voff], (__m128)(iterv + onesv));
					}
				} else
				  for (off=vx*vlen+w*y; off<(vx+1)*vlen+w*y; off++) { // !vin_bounds
					not_vectorized++;
					if (in_bounds[off] && iter[off] <= cur_iter) {
						zr2 = zr[off]*zr[off];
						zi2 = zi[off]*zi[off];
						if (zi2+zr2 > esc_rad_sq) {
							in_bounds[off] = false;
							if (!(*(int*)(in_bounds+voff)))
								voo_bounds[voff/vlen] = true;
						} else {
							zi[off] = 2*zr[off]*zi[off] + ci[y];
							zr[off] = zr2-zi2 + cr[off-w*y];
							iter[off]++;
						}
					}
				}
			}
		}
	}
	catchup_iter = (catchup_iter > cur_iter) ? catchup_iter : 0;
	return cur_iter++;
}
#endif

template <class vtype>
void MandelModel<vtype>::Draw(SDL_Surface* &surf) {
	memset( draw_this_pixel,	true, sizeof(bool) *w*h);
	this->ReDraw(surf);
}

#ifdef USE_SSE2_and_this_has_been_fixed
template <>
void MandelModel<float>::ReDraw(SDL_Surface* &surf) {
	int not_vectorized = 0; int vectorized=0;
	int x; int vx; int y; int i; int voff; int off;
	v4sf huev; //v4sf satv; v4sf valv; v4sf esc_radv = {esc_rad,esc_rad,esc_rad,esc_rad};
	fvec huefv; fvec satfv; fvec valfv; fvec abszfv;
	//v4sf absz; v4sf absv_sqrt;
	v4si iterv; v4si cur_iterv = {cur_iter,cur_iter,cur_iter,cur_iter}; //v4si onesv = {1,1,1,1}; 
	float absz; float hue; float sat; float val; float esc_rad_inv = 1/esc_rad;
	SDL_LockSurface(surf);
	
#pragma omp parallel for private(x, vx, i, voff, off, iterv, huev, huefv, abszfv, satfv, valfv, hue, sat, val ) schedule(static) 
	for (y=0; y<h; y++) {
		for (vx=0; vx<w/vlen; vx++) {
			voff = vx*vlen+w*y;			
			if (*((int*)( draw_this_pixel+voff ))) {
				iterv = (v4si)_mm_load_ps((float*)&iter[voff]);
				if ((!(*((int*)( draw_this_pixel+voff )) ^ 0x01010101) && (!catchup_iter || 
						!(__builtin_ia32_pmovmskb128(__builtin_ia32_cmpgtps((v4sf)iterv, (v4sf)cur_iterv))))) || redraw_all) {
					vectorized++;
					v4sf zrv = _mm_load_ps(&zr[voff]);
					v4sf ziv = _mm_load_ps(&zi[voff]);
					v4sf abszv = zrv*zrv+ziv*ziv;
					//v4sf absv_sqrt = __builtin_ia32_rcpps(__builtin_ia32_rsqrtps(absv));
					huev = arctan2v(ziv, zrv);
					if (0) {//(vin_bounds[voff]) { // (*((int*)( in_bounds+voff )))
						//vector code here
					} else if (0) {//(voo_bounds[voff]) {
						//vector code here
					} else {
						huefv.v = huev;
						abszfv.v = abszv;
						for (i=0; i<4 ; i++) {
							if (in_bounds[voff+i]) {
								satfv.f[i] = fmodf((1.-esc_rad*abszfv.f[i]), 1);
								valfv.f[i] = fmodf(1 - (abszfv.f[i]*.5),1);
							}
							else {
								fvec absz_sqrtfv;
								absz_sqrtfv.v = __builtin_ia32_rcpps(__builtin_ia32_rsqrtps(abszv));
								satfv.f[i] = 1-(absz_sqrtfv.f[i]-esc_rad)*esc_rad_inv;
								//satfv.f[i] = 1-((sqrtf(abszfv.f[i])-esc_rad)/esc_rad);
								valfv.f[i] = (sinf(sinf(pow(iter[voff+i], .5)*2))*.5+.5);
								draw_this_pixel[voff+i]=false;
							}
							satfv.f[i] = (satfv.f[i] < 0) ? 0 : satfv.f[i];
							valfv.f[i] = (valfv.f[i] < 0) ? 0.01 : valfv.f[i];
							*((Uint32*)surf->pixels+voff+i) = HSVtoRGB(surf->format, huefv.f[i], satfv.f[i], valfv.f[i]);
						}
					}
						
			} else		for (x=vx*4; x<vx*4+4 ; x++) {
				off = x+w*y;
				not_vectorized++;
				if (draw_this_pixel[off] && iter[off] <= cur_iter || redraw_all) {
					absz = zr[off]*zr[off]+zi[off]*zi[off];
					hue = arctan2((float)zi[off],(float)zr[off]);
					if (in_bounds[off]) {
						sat = fmodf((1.-esc_rad*absz), 1);
						val = fmodf(1 - (absz*.5),1);
					}
					else {
						sat = 1-((sqrtf(absz)-esc_rad)/esc_rad);
						val = (sinf(sinf(pow(iter[off], .5)*2))*.5+.5) ;
						draw_this_pixel[off]=false;
					}
					sat = (sat < 0) ? 0 : sat;
					val = (val < 0) ? 0 : val;
					*((Uint32*)surf->pixels+off) = HSVtoRGB(surf->format, hue, sat, val);
					}
				}
			}
		}
	}
	redraw_all = false;
	SDL_UnlockSurface(surf);
	SDL_UpdateRect(surf,0,0,0,0);
	//printf(", nvd = %6i; vd = %6i", not_vectorized, vectorized);
}
				
#endif
template <class vtype>
void MandelModel<vtype>::ReDraw(SDL_Surface* &surf) {
	int x; int y; int vx; int yw; int off;
	float hue; float sat; float val;
	float absz; float absz_sqrt;
//	float esc_rad_inv = 1/esc_rad;
//	float esc_rad_inv_sq = esc_rad_inv*esc_rad_inv;
	SDL_LockSurface(surf);
#pragma omp parallel for private(x, vx, yw, off, hue, sat, val, absz, absz_sqrt ) schedule(static) 
	for (y=0; y<h; y++) {
		yw = y*w;
		for (vx=0;vx<w/4;vx++) if (*(int*)(draw_this_pixel + vx*4+yw))
		for (x=vx*4; x<vx*4+4 ; x++) {
			off = x+yw;
			if ((draw_this_pixel[off] && iter[off] <= cur_iter) || redraw_all) {
				absz = zr[off]*zr[off]+zi[off]*zi[off];
				hue = arctan2((float)zi[off],(float)zr[off]);
				if (absz > esc_rad_sq) in_bounds[off] = false;
				if (in_bounds[off]){
					sat = 1.-esc_rad*absz;
					val = 1.-.5*absz;
//					sat = fmodf((1.-esc_rad*absz), 1);
//					val = fmodf(1 - (absz*.5),1);
					//sat = 1-.5* zr[off] * esc_rad_inv;
					//val = .5;//1-.5* zi[off] * esc_rad_inv;
					
					//sat = (sat < 0) ? 0 : sat;
					//val = (val < 0) ? 0 : val;
				}
				else {
					sat = 1-((sqrtf(absz)-esc_rad)/esc_rad);
					val = (sinf(sinf(pow(iter[off], .5)*2))*.5+.5) ;
					draw_this_pixel[off]=false;
				}
				*((Uint32*)surf->pixels+off) = HSVtoRGB(surf->format, hue, sat, val);
			}
		}
	}
	redraw_all = false;
	SDL_UnlockSurface(surf);
	SDL_UpdateRect(surf,0,0,0,0);
}

template <class vtype>
int MandelModel<vtype>::GetVisIter(int max_iter) {
	int vis_iter = max_iter;
	for (int i=0; i<w*h; i++)
		if (!in_bounds[i] && iter[i] < vis_iter) 
			vis_iter = iter[i];
	
	int y=0; int x=0; int i;
	vtype zr2; vtype zi2; vtype z_r; vtype z_i;
	while (y<h) {
		z_r = zr[x+w*y];
		z_i = zi[x+w*y];
		for (i=iter[x+w*y]; i<vis_iter; i++) {
			zr2 = z_r*z_r;
			zi2 = z_i*z_i;
			if (zi2+zr2 > esc_rad_sq) {
				//in_bounds[x+w*y] = false;
				break;
			}
			z_i = 2*z_r*z_i + ci[y];
			z_r = zr2 - zi2 + cr[x];
		}
		vis_iter = (i < vis_iter) ? i : vis_iter;
		if (y==0 || y==h-1) {if (x < w-1) {x = (x+8 >= w) ? w-1 : x+8;} else {x = 0; y+=8;}}
		else {if (x == 0) {x = w-1;} else {x = 0; y = (y+8 >= h) ? h-1 : y+8;}}
	}
	vis_iter = (vis_iter == max_iter) ? 0 : vis_iter;
	return vis_iter;
}

template <class vtype>
void MandelModel<vtype>::ChangeColors() {
	HSVtoRGBindex = (HSVtoRGBindex + 1) % HSVtoRGBsize;
	HSVtoRGB = HSVtoRGBs[HSVtoRGBindex];
	return;
}

template <class vtype>
void MandelModel<vtype>::SetColors(int i) {
	HSVtoRGBindex = i % HSVtoRGBsize;
	HSVtoRGB = HSVtoRGBs[HSVtoRGBindex];
	return;
}

template <class vtype>
Position MandelModel<vtype>::GetPos() {
	struct Position pos;
	pos.rc=rc; pos.ic=ic; pos.rs=rs; pos.is=is;
	return pos;
}


/*#define run_a_step tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp
 run_a_step; run_a_step; run_a_step; run_a_step; 
 run_a_step; run_a_step; run_a_step; run_a_step;
 run_a_step; run_a_step; run_a_step; run_a_step; 
 run_a_step; run_a_step; run_a_step; 
 
 vtype zr2 = zrt*zrt;
 vtype zi2 = zit*zit;
 if (zi2+zr2 > esc_rad_sq) {
 printf("\nGuessed wrong during prejump at x=%i, y=%i, i=%i, cx=%f, cy=%f\n", x,y,i,cx,cy);
 cont = false;
 } else {
 zi[off] = 2*zr[off]*zi[off] + cy;
 zr[off] = zr2-zi2 + cx;
 i = i + step_size;
 iter[off] = i;
 }*/

/*				while (cont && i<target_iter) { // pre-jump 2
 tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
 tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
 tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
 tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;	tmp=zrt*zrt-zit*zit+cx;  zit=2*zrt*zit+cy;  zrt=tmp;
 
 if (zrt*zrt+zit*zit < esc_rad_sq) {
 zr[x] = zrt; zi[x] = zit; i = i + step_size/2; iter[off] = i;
 } else {
 cont = false;
 }
 }*/
