/*
 *  model.h
 *  rima
 *
 *  Created by Jonathan Toomim on 8/2/08.
 *  Copyleft 2008. All rights reversed.
 *
 */
typedef Uint32 (*colorconvtype)(SDL_PixelFormat *, float, float, float);
struct Position {
	long double rc;
	long double ic;
	long double rs;
	long double is;
};

template <class vtype>
class MandelModel {
public:
	//MandelModel() {MandelModel(640,480,0,0,2,2);};
	MandelModel(int xsize, int ysize, long double rcenter, long double icenter, long double rsize, long double isize);
	~MandelModel();
	void	Zoom(float zoomfac);
	void	Pan(float xpan, float ypan);
	void	Clear();
	void	CatchUp() {this->Jump(catchup_iter);};
	virtual void	Jump(int target_iter);
	virtual int 	Step();
	virtual void	Draw(SDL_Surface* &surf);
	virtual void	ReDraw(SDL_Surface* &surf);
	int		GetVisIter(int max_iter=10000);
	int 	GetCurIter() {return cur_iter;};
	void	ChangeColors();
	void	SetColors(int i);
	int		GetColorID() {return HSVtoRGBindex;};
	struct Position GetPos();
protected:
	int w; int h; // width and height
	int cur_iter;
	int catchup_iter;
	long double rc;long double ic;long double rs;long double is; /* real and imaginary center and size */
	vtype* zr; vtype* zi; vtype* cr; vtype* ci;
	bool* in_bounds;
	bool* draw_this_pixel;
	bool redraw_all;
	bool screen_door_effect;
	int* iter;
	vtype esc_rad;
	vtype esc_rad_sq;	
	colorconvtype* HSVtoRGBs;
	colorconvtype HSVtoRGB;
	int HSVtoRGBindex;
	int HSVtoRGBsize;
#ifdef __GNUC__
	int vlen;
	bool* vin_bounds;
	bool* voo_bounds;
	bool* vdraw_this_pixel;
#endif
};

class AbstractModel {
	// This class automatically provides an interface to MandelModel that doesn't
	// depend on what precision level we're at, allowing for transparent precision
	// adjustment
public:
	AbstractModel(int xsize, int ysize, struct Position newpos);
	~AbstractModel();
	void	Zoom(float zoomfac);
	void	Pan(float xpan, float ypan);
	void	Clear();
	void	CatchUp();
	void	Jump(int target_iter);
	int 	Step();
	void	Draw(SDL_Surface* &surf);
	void	ReDraw(SDL_Surface* &surf);
	int		GetVisIter(int max_iter=10000);
	int 	GetCurIter();
	void	ChangeColors();
	void	SetColors(int i);
	int		GetColorID();
	struct	Position GetPos();
	char	GetPrecision() {return prec;};
	char	GetPrecisionRequirement(int x, int y, struct Position newpos);
	void	SetPrecision(char p='0');
	void	SetPrecision(char p, int xsize, int ysize);
protected:
	MandelModel<float> *fmodel;
	MandelModel<double> *dmodel;
	MandelModel<long double> *lmodel;
	char prec;
	int fprec;
	int dprec;
	int lprec;
	int w; 
	int h; // width and height
	struct Position pos; /* real and imaginary center and size */
};
